from autogalaxy.aggregator.galaxies import GalaxiesAgg
from autogalaxy.aggregator.imaging.imaging import ImagingAgg
from autogalaxy.aggregator.imaging.fit_imaging import FitImagingAgg
from autogalaxy.aggregator.interferometer.interferometer import InterferometerAgg
from autogalaxy.aggregator.interferometer.fit_interferometer import FitInterferometerAgg
from autogalaxy.aggregator.ellipse.ellipses import EllipsesAgg
from autogalaxy.aggregator.ellipse.multipoles import MultipolesAgg
from autogalaxy.aggregator.ellipse.fit_ellipse import FitEllipseAgg
